/*
 *  BINtegrity aka Syswall
 *
 */

#undef __KERNEL__
#define __KERNEL__
 
#undef MODULE
#define MODULE

#include <linux/mm.h>
#include <linux/fs.h>
#include <linux/elf.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/file.h>
#include <linux/mman.h>
#include <linux/fcntl.h>
#include <linux/ptrace.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/vmalloc.h>
#include <linux/kallsyms.h>
#include <linux/syscalls.h>
#include <linux/fs_struct.h>
#include <linux/jiffies.h>

#include "syswall.h"
#include "hash.h"

#ifdef CONFIG_ARM
#define ELF_BASE 0x8000
#include "syswall_arm.h"
#endif
#ifdef CONFIG_MIPS
#include "syswall_mips.h"
#define ELF_BASE 0x400000
#endif
#ifdef CONFIG_X86_32
#error x86 currently not supported
#include "syswall_x86.h"
#define ELF_BASE 0x8048000
#endif

#define COLLECT_STATS 1
#define COLLISION_FREE 1

const char *elf_magic = "\x7f\x45\x4c\x46";

MODULE_LICENSE("Dual BSD/GPL");

unsigned **sys_call_table;
#define HASHMAP_MM (4096 * 6)


static unsigned long **get_sys_call_table(void) {
#ifdef CONFIG_MIPS
    return 0x8006a770;
#else
	unsigned long int offset = PAGE_OFFSET;
	unsigned long **sct;
    
	while (offset < ULLONG_MAX) {
		sct = (unsigned long **)offset;
		if (sct[__NR_close] == (unsigned long *) sys_close)
			return sct;
		offset += sizeof(void *);
	}
	return NULL;
#endif
}

int parse_elf_file(struct file *elf_file, elf_info *ei, unsigned char **elf_mem) {
    size_t elf_size = i_size_read(elf_file->f_path.dentry->d_inode);
    size_t first_chunk_size = min(elf_size, ELF_READ_SIZE);

    *elf_mem = vmalloc(first_chunk_size);
    if (*(elf_mem))
        return -2;
    
    if (!kernel_read(elf_file, 0, *elf_mem, first_chunk_size)) {
        return -2;
    }

    unsigned char *mem = *elf_mem;

    for (unsigned i = 0; elf_magic[i] != '\0' && i < 0x4; i++) {
        if (elf_magic[i] != mem[i]) {
            return -1;
        }
    }
    
    Elf32_Ehdr *elf_hdr = (Elf32_Ehdr *)mem;
    ei->type = elf_hdr->e_type;

    Elf32_Phdr *p_hdr = NULL;
    unsigned dyn_offset = 0;
    for (unsigned i = 0; i < elf_hdr->e_phnum; i++) {
        p_hdr = (Elf32_Phdr *)(mem + elf_hdr->e_phoff + i * elf_hdr->e_phentsize);
        if (p_hdr->p_type == PT_DYNAMIC) {
            dyn_offset = p_hdr->p_offset;
            break;
        }
    }

    if (!dyn_offset) {
        return -2;
    }

    unsigned char* dyn_mem = mem;
    if (dyn_offset > first_chunk_size) {
        dyn_mem = kmalloc(0xd8, GFP_KERNEL);
        if (!kernel_read(elf_file, dyn_offset, dyn_mem, 0xd8)) {
            kfree(dyn_mem);
            return -2;
        }
        dyn_offset = 0;
    }

    Elf32_Dyn *dyn = dyn_mem + dyn_offset;
    Elf32_Sym *dynsym = NULL;
    unsigned *dynsym_entries = NULL;
    char *dynstr = NULL;
    while (dyn->d_tag != DT_NULL) {
        if (dyn->d_tag == DT_SYMTAB)
            dynsym = dyn->d_un.d_ptr;
        if (dyn->d_tag == DT_STRTAB) 
            dynstr = dyn->d_un.d_ptr;
        if (dyn->d_tag == DT_HASH) {
            dynsym_entries = (unsigned *)dyn->d_un.d_ptr;
            dynsym_entries++;
        }
        dyn++;
    }

    if (!dyn_offset)
        kfree(dyn_mem);

    if (ei->type == ET_DYN) {
        ei->dynsym = (Elf32_Sym *)((unsigned)dynsym + (unsigned)mem);
        ei->dynstr = (char *)((unsigned)dynstr + (unsigned)mem);
        ei->dynsym_entries =  *(unsigned *)((unsigned)dynsym_entries + (unsigned)mem);
    }
    if (ei->type == ET_EXEC) {
        ei->dynsym = (Elf32_Sym *)((unsigned)dynsym - ELF_BASE + (unsigned)mem);
        ei->dynstr = (char *)((unsigned)dynstr - ELF_BASE + (unsigned)mem);
        ei->dynsym_entries =  *(unsigned *)((unsigned)dynsym_entries - ELF_BASE + (unsigned)mem);
    }

    if ((unsigned)ei->dynsym - (unsigned)(mem) > first_chunk_size || (unsigned)ei->dynstr - (unsigned)(mem) > first_chunk_size) {
        return -3;
    }

    return 0;
}

int is_export(elf_info *ei, unsigned address, char **export_sym)
{
    unsigned char esym_cnt = 0;
    for (unsigned j = 0; j < ei->dynsym_entries && esym_cnt < 3; j++) {
        if (ei->type == ET_DYN && ((ELF32_ST_BIND(ei->dynsym[j].st_info) == STB_GLOBAL) || (ELF32_ST_BIND(ei->dynsym[j].st_info) == STB_WEAK)) && (ELF32_ST_TYPE(ei->dynsym[j].st_info) == STT_FUNC) && ei->dynsym[j].st_shndx != SHN_UNDEF && ei->dynsym[j].st_value <= address && address < ei->dynsym[j].st_value + ei->dynsym[j].st_size) {
            export_sym[esym_cnt] = ei->dynstr + ei->dynsym[j].st_name;
            esym_cnt++;
        }
    }
    
    if (!esym_cnt)
        return -1;
    
    return 0;
}

int is_import(elf_info *ei, char **export_sym)
{
    Elf32_Sym *dynsym = ei->dynsym;

    if (!export_sym[0]) {
        WARN("export_sym[0] == 0\n");
        return 0;
    }

    for (unsigned j = 0; j < ei->dynsym_entries; j++) {
        if ((ELF32_ST_BIND(ei->dynsym[j].st_info) == STB_GLOBAL) && (ELF32_ST_TYPE(ei->dynsym[j].st_info) == STT_FUNC) && ei->dynsym[j].st_shndx == SHN_UNDEF) {
            if (!strcmp(export_sym[0], ei->dynstr + ei->dynsym[j].st_name))
                return 0;
            if (export_sym[1] && !strcmp(export_sym[1], ei->dynstr + ei->dynsym[j].st_name))
                return 0;
            if (export_sym[2] && !strcmp(export_sym[2], ei->dynstr + ei->dynsym[j].st_name))
                return 0;
        }
    }
    return -1;
}

int vet_symbol(struct vm_area_fix_t *vma, unsigned vma_entries, unsigned export_address, unsigned import_reference) 
{
    char *export_sym[3] = {NULL, NULL, NULL};
    unsigned char *export_elf_mem;
    for (unsigned i = 0; i < vma_entries; i++) {
        if (vma[i].start <= export_address && export_address < vma[i].end) {
            elf_info ei;
            if (parse_elf_file(vma[i].f, &ei, &export_elf_mem) != 0) {
            }
            else if (is_export(&ei, export_address - vma[i].start, export_sym) == 0) {
                DPRINT("\t[VS]\tAddress is in exported function %s\n", export_sym[0]);
                break;
            }
            vfree(export_elf_mem);
            export_elf_mem = NULL;
            break;
        }
    }

    if (!export_sym[0]) {
        DPRINT("\t[VS]\tAddress not exported\n");
        vfree(export_elf_mem);
        return 0;
    }


// start: this need to be adjusted for your >>>specific<<< platform, if this does not match you will have issues

#ifdef CONFIG_MIPS
    int stop = 0;
    if (!strcmp(export_sym[0], "_dl_load_elf_shared_library")) {
        export_sym[1] = "_dl_load_shared_library";
    } else if (!strcmp(export_sym[0], "tcgetattr")) {
        stop = 1;
    } else if (!strcmp(export_sym[0], "execvp")) {
        stop = 1;
    }
    if (stop) {
        vfree(export_elf_mem);
        return 0;
    }
#endif

#ifdef CONFIG_ARM
    int stop = 0;
    if (!strcmp(export_sym[0], "__cxa_atexit") || !strcmp(export_sym[0], "__atexit_register_cleanup") || !strcmp(export_sym[0], "__cxa_finalize"))
        stop = 1;
    else if (!strcmp(export_sym[0], "mmap"))
        stop = 1;
    else if (!strcmp(export_sym[0], "pthread_create"))
        stop = 1;
    else if (!strcmp(export_sym[0], "cpuacct_add"))
         stop = 1;
    else if (!strcmp(export_sym[0], "ioctl"))
        stop = 1;
    else if (!strcmp(export_sym[0], "__res_nsend"))
        stop = 1;
    else if (!strcmp(export_sym[0], "socket_local_client_connect"))
        stop = 1;
    else if (!strcmp(export_sym[0], "getaddrinfo"))
        stop = 1;
    else if (!strcmp(export_sym[0], "__system_property_set"))
        stop = 1;
    else if (!strncmp(export_sym[0], "_Z", 2))
        stop = 1;
    if (stop) {
        vfree(export_elf_mem);
        return 0;
    }
#endif

// end platform specifics!

    unsigned found = 0;
    for (unsigned i = 0; i < vma_entries && !found; i++) {
        if (!import_reference || (vma[i].start <= import_reference && import_reference < vma[i].end)) {
            elf_info ei;
            unsigned char *elf_mem;
            if (parse_elf_file(vma[i].f, &ei, &elf_mem) != 0) {
                DPRINT("\t[VS]\tCould not parse elf file\n");
                vfree(elf_mem);
                continue;
            }
            if (is_import(&ei, export_sym) == 0) {
                DPRINT("\t[VS]\tFunction is imported by ELF at %#x\n", vma[i].start);
                found = 1;
            }
            if (!found) {
                WARN("\t[VS]\tImport not found: %s\n", export_sym[0]);
            }
            vfree(elf_mem);
        }
    }

    if (export_elf_mem)
        vfree(export_elf_mem);

    if (!found) {
        return -1;
        //return 0;
    } else {
        return 0;
    }
}

int check_exit()
{
    struct hashmap_t *hm = 0;
    struct vm_area_struct *vm_area = 0;
    
    if (current->mm && current->mm->mmap)
        vm_area = current->mm->mmap;

    if (vm_area && vm_area->vm_start == 0x1000) {
        hm = (struct hashmap_t*) vm_area->vm_start;
        vm_area = vm_area->vm_next;

        unsigned int used = 0;
        unsigned int args = 0;
        unsigned int vetsym = 0;
        struct vetmap_t *stats = (struct vetmap_t*) hm->hm;
        printk("[STATS] stats for %d (%s) jiffies: ", current->pid, current->comm);
        for (int i = 0; i < HASHMAP_SIZE; i++) {
            if (stats[i].addr != 0) {
                used++;
                if (stats[i].subreturn_valid)
                    vetsym++;
                printk("TS = %lu ", jiffies_to_msecs(stats[i].ts));
                for (int i2 = 0; i2 < 4; i2++)
                    if (stats[i].argv[i2] != 0 && stats[i].argv[i2] != 0xdeadbeef)
                        args++;
            }
            //if (stats[i].over)
            //    printk("[STATS] BINtegrity stats for %d (%s) bucket[%d] was overwritten %d times\n", current->pid, current->comm, i, stats[i].over);
        }
        printk(" TSE = %lu\n", jiffies_to_msecs(jiffies));
        printk("[STATS] stats for %d (%s) used buckets %d and %d args vmas = %d vetsym = %d\n", current->pid, current->comm, used, args, hm->vma_num, vetsym);
    }
    else {
        //printk("[STATS] stats for %d (%s) no stats!\n", current->pid, current->comm);
    }

    return 0;
}

int add_tcb(void *addr) { }

int add_single_vma(struct vm_area_fix_t *vma, unsigned int *vma_num, struct vm_area_struct *vm_area) {
    if (vm_area->vm_file && (vm_area->vm_flags & VM_EXECUTABLE || vm_area->vm_flags & VM_EXEC)) {
        DPRINT("\t[VMA] = %d \tadding VMA at %#x for file %s\n", *vma_num, vm_area->vm_start, vm_area->vm_file->f_path.dentry->d_name.name);
        vma[*vma_num].f = vm_area->vm_file;
        vma[*vma_num].start = vm_area->vm_start;
        vma[*vma_num].end = vm_area->vm_end;
        vma[*vma_num].dynamic = 1;
        (*vma_num)++;
    }
}

int fill_vma(struct vm_area_fix_t *vma, unsigned int *vma_num, struct vm_area_struct *vm_area)
{
    unsigned int vma_c = 0;

    do {
        if (vm_area->vm_file && (vm_area->vm_flags & VM_EXECUTABLE || vm_area->vm_flags & VM_EXEC)) {
            DPRINT("\t[VMA] = %d \tadding VMA at %#x for file %s\n", vma_c, vm_area->vm_start, vm_area->vm_file->f_path.dentry->d_name.name);
            vma[vma_c].f = vm_area->vm_file;
            vma[vma_c].start = vm_area->vm_start;
            vma[vma_c].end = vm_area->vm_end;
            vma[vma_c].dynamic = 0;
            vma_c++;
        }
        vm_area = vm_area->vm_next;
    } while (vm_area != NULL);

    *vma_num = vma_c;

    return 1;
}

#define CL ORIGIN
#include "check_syscall_template.c"
#undef CL

#define CL ARGUMENTS
#include "check_syscall_template.c"
#undef CL

#define CL SYMBOLS
#include "check_syscall_template.c"
#undef CL

static int __init syswall_init(void) {
    sys_call_table = get_sys_call_table();
    if (!sys_call_table) {
        WARN("Syscall table not found\n");
        return -1;
    }

    disable_page_protection();
    #include "hook.h"
    enable_page_protection();
    return 0;
}
 
static void __exit syswall_cleanup(void) {
    if(!sys_call_table)
		return;
    
	disable_page_protection();
    #include "unhook.h"
    enable_page_protection();
    printk(KERN_INFO "Original handlers restored.\n");
}

module_init(syswall_init);
module_exit(syswall_cleanup);
MODULE_AUTHOR("MC");
MODULE_DESCRIPTION("BINtegrity");
