/*
 *  BINtegrity aka Syswall
 *
 */

#include <linux/fs.h>
#include <linux/elf.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/kallsyms.h>
#include <linux/syscalls.h>
#include <linux/uaccess.h>

#include "syswall.h"
#include "mips_dasm.h"

#define LOOKBACK 0x50

void disable_page_protection(void) {}

void enable_page_protection(void) {}

#define DPRINT(...) ;

void check_reg_mod(mips_instruction_t *instruction, unsigned *regmodlist) {
    if (instruction->dst == V0) {
        if (instruction->op == LW && instruction->src[0] == FP && instruction->imm >= 0) {
            *regmodlist &= ~(uint32_t)(MOD_REG_V0);
        } else {
            *regmodlist |= (uint32_t)(MOD_REG_V0);
        }
    }

    if (A0 <= instruction->dst && instruction->dst <= A3) {
        DPRINT("instruction %s dst %s src %s, %s imm %u\n", mips_operation_names[instruction->op], mips_register_names[instruction->dst], mips_register_names[instruction->src[0]], mips_register_names[instruction->src[1]], instruction->imm);
        if (instruction->op == LW) {
            *regmodlist &= ~(1 << (instruction->dst - A0));
        } else if (instruction->op == ADDU && instruction->src[0] == V0 && instruction->src[1] == ZERO && !(*regmodlist & (uint32_t)(MOD_REG_V0))) {
            *regmodlist &= ~(1 << (instruction->dst - A0));
        } else {
            *regmodlist |= (1 << (instruction->dst - A0));
        }
    }
    DPRINT("regmodlist: %#x\n", *regmodlist);
}

int get_args(unsigned char *buffer, unsigned offset, reg_state_t *state, unsigned *regmodlist) {
    int start = 0;
    for (int i = offset; i >= 0; i -= 4) {
        mips_instruction_t instruction;
        uint32_t instr_raw = (buffer[i] << 24) | (buffer[i+1] << 16) | (buffer[i+2]) << 8 | buffer[i+3];
        if(mips_disassemble(&instruction, instr_raw)) {
            if (instruction.type == JUMP) {
                break;
            }
            if (instruction.type == ARTH_LOG && instruction.dst == SP && instruction.src[0] == SP && instruction.imm < 0) {
                break;
            }
        }
        start = i;
    }
    
    for (unsigned i = 0; i < sizeof(state->regs)/4; i++) {state->regs[i] = 0xdeadbeef;}
    state->regs[ZERO] = 0;
    for (int i = start; i < LOOKBACK; i += 4) {
        mips_instruction_t instruction;
        uint32_t instr_raw = (buffer[i] << 24) | (buffer[i+1] << 16) | (buffer[i+2]) << 8 | buffer[i+3];
        if (mips_disassemble(&instruction, instr_raw)) {
            int ret = mips_execute(&instruction, state);
            if (ret == -1) {
                DPRINT("\t\tInstruction [%#x] execution of %s not supported\n", i, mips_operation_names[instruction.op]);
            } else if (ret == -2) {
                DPRINT("\t\tInstruction operand unknown\n");
            }
            if (regmodlist) {
                check_reg_mod(&instruction, regmodlist);
            }
        } else {
            DPRINT("\t\tInvalid instruction\n");
        }
    }
}

int vet_syscall(struct file *f, unsigned offset, unsigned syscall_number, unsigned int *regmodlist) {
    unsigned char *buffer = kmalloc(LOOKBACK, GFP_KERNEL);
    kernel_read(f, offset - LOOKBACK, buffer, LOOKBACK);

    int retval = -2;

    mips_instruction_t instruction;
    uint32_t instr_raw = (buffer[LOOKBACK-4] << 24) | (buffer[LOOKBACK-3] << 16) | (buffer[LOOKBACK-2]) << 8 | buffer[LOOKBACK-1];
    mips_disassemble(&instruction, instr_raw);
    if (instruction.op == SYSCALL) {
        retval = -1;
    } else {
        kfree(buffer);
        return retval;
    }

    reg_state_t state;
    int r = get_args(buffer, LOOKBACK-0x8, &state, regmodlist);
    DPRINT("\t\tArguments: A0:%#x A1:%#x A2:%#x V0:%#x\n", state.regs[A0], state.regs[A1], state.regs[A2], state.regs[V0]);
    
    if (offset == 0x16e90) {
        *regmodlist |= (uint32_t)(MOD_REG_A3);
    }
                
    if (state.regs[V0] == syscall_number) {
        retval = 0;
    }

    kfree(buffer);
    
    return retval;
}

int vet_subreturn(struct file *f, unsigned offset, unsigned *argv, unsigned argc, unsigned int regmodlist, struct vetmap_t **vm) {
    unsigned char *buffer = kmalloc(LOOKBACK, GFP_KERNEL);
    kernel_read(f, offset - LOOKBACK, buffer, LOOKBACK);
    
    int retval = -3;

    mips_instruction_t instruction;
    uint32_t instr_raw = (buffer[LOOKBACK-8] << 24) | (buffer[LOOKBACK-7] << 16) | (buffer[LOOKBACK-6]) << 8 | buffer[LOOKBACK-5];
    mips_disassemble(&instruction, instr_raw);
    if (instruction.type == JUMP) {
        if  (instruction.op == JAL || instruction.op == BGEZAL || instruction.op == BLTZAL) {
            retval = 0;
        } else {
            retval = -2;
        }
    } else {
        //WARN("\t\tNot a jump: %s\n", mips_operation_names[instruction.op]);
    }

    reg_state_t state;
    if (retval != -3) {

        int r = get_args(buffer, LOOKBACK-0xc, &state, NULL);
        
        if (retval == -2) {
            if (state.regs[instruction.src[0]] != 0xdeadbeef) {
                retval = 0;
            } else {
            }
        }

        if (retval == 0) {
            DPRINT("\t\tArguments: A0:%#x A1:%#x A2:%#x A3:%#x\n", state.regs[A0], state.regs[A1], state.regs[A2], state.regs[A3]);
            DPRINT("\t\tModified registers: %#x\n", regmodlist);
            for (unsigned i = 0; i < argc; i++) {
                if (regmodlist & (1 << i)) {
                    (*vm)->argv[i] = 0xdeadbeef;
                    continue;
                }
                if (state.regs[A0 + i] != 0xdeadbeef && state.regs[A0 + i] != argv[i]) {
                    DPRINT("\t\tArgument mismatch in A%u\n", i);
                    retval = -1;
                    break;
                } else {
                    (*vm)->argv[i] = state.regs[A0 + i];
                }
            }
        }
    }

    kfree(buffer);
    
    return retval;
}
